<?php

namespace Security;

use Exception;
use Laravel\Config;
use Laravel\Lang;
use Laravel\Request;
use Laravel\Database\Manager as DB;

/**
 * Access Control List class.
 *
 * The purpose of this class is to provide access control to RESTful
 * URIs. The database setup script can be found in the config file
 * named acl.php.
 *
 * @author Amir Khawaja <khawaja.amir@gmail.com>
 */
class Acl
{
    private static $_conf;

    /**
     * Determine if a User is in a role.
     * @static
     * @throws \Exception
     * @param string $role
     * @param int $user_id
     * @return bool
     */
    public static function is_user_in_role($role, $user_id = -1)
    {
        $conf = static::config();

        $role = DB::table($conf['table_role'])
            ->join($conf['table_user_role'],
                   $conf['table_user_role'] . '.role_id', '=',
                   $conf['table_role'] . '.id')
            ->join($conf['table_user'],
                   $conf['table_user'] . '.id', '=',
                   $conf['table_user_role'] . '.user_id')
            ->where($conf['table_role'] . '.name', '=', $role)
            ->where($conf['table_user'] . '.id', '=', $user_id)
            ->count();

        return (!is_null($role) && $role > 0);
    }

    /**
     * Check to see if a Role has permission to access a Resource.
     * @static
     * @throws \Exception
     * @param string $user_id
     * @param null|string $resource
     * @return bool
     */
    public static function has_permission($user_id, $resource = null)
    {
        $conf = static::config();

        if (is_null($resource)) {
            $resource = Request::route()->key;
        }

        if (is_null($user_id) || !is_numeric($user_id)) {
            throw new Exception(Lang::line('acl.user_id_missing')->get());
        }

        $perm = DB::table($conf['table_role'])
            ->join($conf['table_resource'],
                   $conf['table_resource'] . '.role_id', '=',
                   $conf['table_role'] . '.id')
            ->join($conf['table_user_role'],
                   $conf['table_user_role'] . '.role_id', '=',
                   $conf['table_role'] . '.id')
            ->join($conf['table_user'],
                   $conf['table_user'] . '.id', '=',
                   $conf['table_user_role'] . '.user_id')
            ->where($conf['table_user'] . '.id', '=', $user_id)
            ->where($conf['table_resource'] . '.uri', '=', $resource)
            ->count();

        return (!is_null($perm) && $perm > 0);
    }

    /**
     * Get the Roles of a given User.
     * @static
     * @throws \Exception
     * @param int $user_id
     * @return null|object
     */
    public static function roles_for($user_id)
    {
        $conf = static::config();

        if (is_null($user_id) || !is_numeric($user_id)) {
            throw new Exception(Lang::line('acl.user_id_missing')->get());
        }

        $roles = DB::table($conf['table_role'])
            ->join($conf['table_user_role'],
                   $conf['table_user_role'] . '.role_id', '=',
                   $conf['table_role'] . '.id')
            ->join($conf['table_user'],
                   $conf['table_user'] . '.id', '=',
                   $conf['table_user_role'] . '.user_id')
            ->where($conf['table_user'] . '.id', '=', $user_id)
            ->get();

        return is_null($roles) ? null : $roles;
    }

    /**
     * Add a new role.
     *
     * A Role is what gets protected. A Role is also what gets assigned to a User.
     *
     * @static
     * @throws \Exception
     * @param $role_name
     * @return int
     */
    public static function add_role($role_name)
    {
        $conf = static::config();

        return DB::table($conf['table_role'])
            ->insert_get_id(array('name' => $role_name), $conf['table_role_seq']);
    }

    /**
     * Add a new Resource for a given Role.
     *
     * A Resource is a URI. Some examples of URI are:
     * - GET /
     * - GET /admin
     * - POST /admin/user/new
     *
     * @static
     * @throws \Exception
     * @param string $role
     * @param array $routes
     * @return void
     */
    public static function add_routes_for($role, $routes)
    {
        if (is_null($routes) || !is_array($routes)) {
            throw new Exception(Lang::line('acl.routes_datatype_mismatch'));
        }

        $conf = static::config();
        $role_id = static::find_role_id($role);

        $len = count($routes);
        for ($i = 0; $i < $len; $i++) {
            DB::table($conf['table_resource'])
                ->insert(array('role_id' => $role_id, 'uri' => trim($routes[$i])));
        }
    }

    /**
     * Get all the unassigned Resources for a given Role.
     * @static
     * @param int|string $role
     * @return null|array
     */
    public static function unassigned_routes_for($role)
    {
        $conf = static::config();
        $role_id = static::find_role_id($role);

        return DB::connection()->query("
            select au.uri
            from $conf[table_app_uri] au
            where au.uri not in
            (
                select r.uri from $conf[table_resource] r
                inner join $conf[table_role] ro on ro.id = r.role_id
                where r.role_id = :role_id
            )", array('role_id' => $role_id));
    }

    /**
     * Get all the Resources assigned to a Role.
     * @static
     * @param $role
     * @return null|array
     */
    public static function assigned_routes_for($role)
    {
        $conf = static::config();
        $role_id = static::find_role_id($role);

        return DB::query("
            select r.id, r.uri from $conf[table_resource] r
            inner join $conf[table_role] ro on ro.id = r.role_id
            where r.role_id = :role_id
        ", array('role_id' => $role_id));
    }

    /**
     * Remove access to Resources for a given Role.
     * @static
     * @throws \Exception
     * @param int|string $role
     * @param array $routes
     * @return void
     */
    public static function remove_routes_for($role, $routes = array())
    {
        if (is_null($role)) {
            throw new Exception(Lang::line('acl.role_missing'));
        }

        if (is_null($routes) || !is_array($routes)) {
            throw new Exception(Lang::line('acl.routes_datatype_mismatch'));
        }

        $conf = static::config();
        $role_id = static::find_role_id($role);

        $len = count($routes);
        for ($i = 0; $i < $len; $i++) {
            DB::table($conf['table_resource'])
                ->where('role_id', '=', $role_id)
                ->where('id', '=', $routes[$i])
                ->delete();
        }
    }

    /**
     * Refresh the App URI table.
     * @static
     * @return void
     */
    public static function refresh_app_uri()
    {
        $routes = Helper::list_all_uri();

        // Delete all the records first.
        DB::query('delete from app_uri');

        // Add the routes in the application.
        foreach ($routes as $uri) {
            DB::table('app_uri')->insert(array('uri' => trim($uri)));
        }
    }

    /**
     * Get the configuration for ACL.
     * @static
     * @throws \Exception
     * @return array
     */
    protected static function config()
    {
        if (is_null(static::$_conf)) {
            static::$_conf = Config::get('acl');

            if (is_null(static::$_conf)) {
                throw new Exception(Lang::line('acl.config_file_missing')->get());
            }
        }

        return static::$_conf;
    }

    /**
     * Find the Role ID of a given role.
     * @static
     * @throws \Exception
     * @param int|string $role
     * @return int
     */
    protected static function find_role_id($role)
    {
        $conf = static::config();

        // If the role parameter is a number, assume it is the Role ID. Otherwise, find the
        // Role ID.
        if (is_numeric($role)) {
            $role_id = $role;
        } else {
            $role_obj = DB::table($conf['table_role'])->where('name', '=', $role)->first();

            if (is_null($role_obj)) {
                throw new Exception(Lang::line('acl.role_does_not_exist', array('role' => $role)));
            }

            $role_id = $role_obj->id;
        }

        return intval($role_id);
    }
}
